@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=2 sw=2 et tw=78: */
 /* ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
  *
@@ -76,10 +77,15 @@
 #include "nsUnicharUtils.h"
 #include "nsCSSFrameConstructor.h"
 #include "nsIBoxLayout.h"
+#include "nsIEventQueueService.h"
+#include "nsIServiceManager.h"
+#include "nsIReflowCallback.h"
 #ifdef XP_WIN
 #include "nsISound.h"
 #endif
 
+static NS_DEFINE_CID(kEventQueueServiceCID, NS_EVENTQUEUESERVICE_CID);
+
 const PRInt32 kMaxZ = 0x7fffffff; //XXX: Shouldn't there be a define somewhere for MaxInt for PRInt32
 
 
@@ -154,7 +160,8 @@ NS_INTERFACE_MAP_END_INHERITING(nsBoxFrame)
 //
 nsMenuPopupFrame::nsMenuPopupFrame(nsIPresShell* aShell)
   :nsBoxFrame(aShell), mCurrentMenu(nsnull), mTimerMenu(nsnull), mCloseTimer(nsnull),
-    mMenuCanOverlapOSBar(PR_FALSE), mShouldAutoPosition(PR_TRUE), mShouldRollup(PR_TRUE)
+    mMenuCanOverlapOSBar(PR_FALSE), mShouldAutoPosition(PR_TRUE), mShouldRollup(PR_TRUE),
+    mInContentShell(PR_TRUE)
 {
   SetIsContextMenu(PR_FALSE);   // we're not a context menu by default
 } // ctor
@@ -206,6 +213,13 @@ nsMenuPopupFrame::Init(nsPresContext*  aPresContext,
   // is fixed.
   viewManager->SetViewContentTransparency(ourView, PR_FALSE);
 
+  nsCOMPtr<nsISupports> cont = aPresContext->GetContainer();
+  nsCOMPtr<nsIDocShellTreeItem> dsti = do_QueryInterface(cont);
+  PRInt32 type = -1;
+  if (dsti && NS_SUCCEEDED(dsti->GetItemType(&type)) &&
+      type == nsIDocShellTreeItem::typeChrome)
+    mInContentShell = PR_FALSE;
+
   // Create a widget for ourselves.
   nsWidgetInitData widgetData;
   widgetData.mWindowType = eWindowType_popup;
@@ -480,10 +494,11 @@ nsMenuPopupFrame::AdjustClientXYForNestedDocuments ( nsIDOMXULDocument* inPopupD
   // nsXULPopupListener).
 
   nsCOMPtr<nsIDOMNode> targetNode;
+  nsCOMPtr<nsIDOMXULDocument2> doc = do_QueryInterface(inPopupDoc);
   if (mContent->Tag() == nsXULAtoms::tooltip)
-    inPopupDoc->GetTooltipNode(getter_AddRefs(targetNode));
+    doc->TrustedGetTooltipNode(getter_AddRefs(targetNode));
   else
-    inPopupDoc->GetPopupNode(getter_AddRefs(targetNode));
+    doc->TrustedGetPopupNode(getter_AddRefs(targetNode));
 
   //NS_WARN_IF_FALSE(targetNode, "no popup/tooltip node on document!");
   nsCOMPtr<nsIContent> targetAsContent ( do_QueryInterface(targetNode) );
@@ -731,7 +746,35 @@ nsMenuPopupFrame::MovePopupToOtherSideOfParent ( PRBool inFlushAboveBelow, PRInt
 
 } // MovePopupToOtherSideOfParent
 
+class nsASyncMenuActivation : public nsIReflowCallback
+{
+public:
+  nsASyncMenuActivation(nsIContent* aContent)
+    : mContent(aContent)
+  {
+  }
 
+  NS_DECL_ISUPPORTS
+
+  NS_IMETHOD ReflowFinished(nsIPresShell* aShell, PRBool* aFlushFlag) {
+    nsAutoString shouldDisplay, menuActive;
+    mContent->GetAttr(kNameSpaceID_None, nsXULAtoms::menuactive, menuActive);
+    if (!menuActive.EqualsLiteral("true")) {
+      mContent->GetAttr(kNameSpaceID_None, nsXULAtoms::menutobedisplayed,
+                        shouldDisplay);
+      if(shouldDisplay.EqualsLiteral("true")) {
+        mContent->SetAttr(kNameSpaceID_None, nsXULAtoms::menuactive,
+                          NS_LITERAL_STRING("true"), PR_TRUE);
+        *aFlushFlag = PR_TRUE;
+      }
+    }
+    return NS_OK;
+  }
+
+  nsCOMPtr<nsIContent> mContent;
+};
+
+NS_IMPL_ISUPPORTS1(nsASyncMenuActivation, nsIReflowCallback)
 
 nsresult 
 nsMenuPopupFrame::SyncViewWithFrame(nsPresContext* aPresContext,
@@ -743,7 +786,7 @@ nsMenuPopupFrame::SyncViewWithFrame(nsPresContext* aPresContext,
   NS_ENSURE_ARG(aPresContext);
   NS_ENSURE_ARG(aFrame);
 
-  if (!mShouldAutoPosition) 
+  if (!mShouldAutoPosition && !mInContentShell)
     return NS_OK;
 
   // |containingView|
@@ -862,9 +905,21 @@ nsMenuPopupFrame::SyncViewWithFrame(nsPresContext* aPresContext,
   // keep 3px margin to the right and bottom of the screen for WinXP dropshadow
   screenWidth -= 3;
   screenHeight -= 3;
+
+  // for content shells, clip to the client area rather than the screen area
+  if (mInContentShell) {
+    nsRect screenRect(screenLeft, screenTop, screenWidth, screenHeight);
+    nsRect rootScreenRect = presShell->GetRootFrame()->GetScreenRect();
+    screenRect.IntersectRect(screenRect, rootScreenRect);
+    screenLeft = screenRect.x;
+    screenTop = screenRect.y;
+    screenWidth = screenRect.width;
+    screenHeight = screenRect.height;
+  }
+
   screenRight = screenLeft + screenWidth;
   screenBottom = screenTop + screenHeight;
-  
+
   PRInt32 screenTopTwips    = NSIntPixelsToTwips(screenTop, p2t);
   PRInt32 screenLeftTwips   = NSIntPixelsToTwips(screenLeft, p2t);
   PRInt32 screenWidthTwips  = NSIntPixelsToTwips(screenWidth, p2t);
@@ -1114,8 +1169,11 @@ nsMenuPopupFrame::SyncViewWithFrame(nsPresContext* aPresContext,
   mContent->GetAttr(kNameSpaceID_None, nsXULAtoms::menuactive, menuActive);
   if (!menuActive.EqualsLiteral("true")) {
     mContent->GetAttr(kNameSpaceID_None, nsXULAtoms::menutobedisplayed, shouldDisplay);
-    if ( shouldDisplay.EqualsLiteral("true") )
-      mContent->SetAttr(kNameSpaceID_None, nsXULAtoms::menuactive, NS_LITERAL_STRING("true"), PR_TRUE);
+    if ( shouldDisplay.EqualsLiteral("true") ) {
+      nsCOMPtr<nsIReflowCallback> cb = new nsASyncMenuActivation(mContent);
+      NS_ENSURE_TRUE(cb, NS_ERROR_OUT_OF_MEMORY);
+      mPresContext->PresShell()->PostReflowCallback(cb);
+    }
   }
 
   return NS_OK;
@@ -1459,16 +1517,10 @@ nsMenuPopupFrame::Enter()
 nsIMenuParent*
 nsMenuPopupFrame::GetContextMenu()
 {
-  if (mIsContextMenu || !nsMenuFrame::sDismissalListener)
+  if (mIsContextMenu)
     return nsnull;
 
-  nsIMenuParent *menuParent = nsMenuFrame::sDismissalListener->GetCurrentMenuParent();
-  if (!menuParent)
-    return nsnull;
-
-  PRBool isContextMenu;
-  menuParent->GetIsContextMenu(isContextMenu);
-  return isContextMenu ? menuParent : nsnull;
+  return nsMenuFrame::GetContextMenu();
 }
 
 nsIMenuFrame*
@@ -1658,10 +1710,14 @@ nsMenuPopupFrame::ShortcutNavigation(nsIDOMKeyEvent* aKeyEvent, PRBool& aHandled
   nsIMenuFrame* result = FindMenuWithShortcut(aKeyEvent, action);
   if (result) {
     // We got one!
+    nsIFrame* frame = nsnull;
+    CallQueryInterface(result, &frame);
+    nsWeakFrame weakResult(frame);
     aHandledFlag = PR_TRUE;
     SetCurrentMenuItem(result);
-    if (action)
+    if (action && weakResult.IsAlive()) {
       result->Enter();
+    }
   }
 
   return NS_OK;
@@ -1697,6 +1753,7 @@ nsMenuPopupFrame::KeyboardNavigation(PRUint32 aKeyCode, PRBool& aHandledFlag)
   PRBool isContainer = PR_FALSE;
   PRBool isOpen = PR_FALSE;
   PRBool isDisabled = PR_FALSE;
+  nsWeakFrame weakFrame(this);
   if (mCurrentMenu) {
     mCurrentMenu->MenuIsContainer(isContainer);
     mCurrentMenu->MenuIsOpen(isOpen);
@@ -1705,13 +1762,19 @@ nsMenuPopupFrame::KeyboardNavigation(PRUint32 aKeyCode, PRBool& aHandledFlag)
     if (isOpen) {
       // Give our child a shot.
       mCurrentMenu->KeyboardNavigation(aKeyCode, aHandledFlag);
+      NS_ENSURE_TRUE(weakFrame.IsAlive(), NS_OK);
     }
     else if (theDirection == eNavigationDirection_End &&
              isContainer && !isDisabled) {
       // The menu is not yet open. Open it and select the first item.
       aHandledFlag = PR_TRUE;
+      nsIFrame* frame = nsnull;
+      CallQueryInterface(mCurrentMenu, &frame);
+      nsWeakFrame weakCurrentFrame(frame);
       mCurrentMenu->OpenMenu(PR_TRUE);
+      NS_ENSURE_TRUE(weakCurrentFrame.IsAlive(), NS_OK);
       mCurrentMenu->SelectFirstItem();
+      NS_ENSURE_TRUE(weakFrame.IsAlive(), NS_OK);
     }
   }
 
@@ -1742,6 +1805,7 @@ nsMenuPopupFrame::KeyboardNavigation(PRUint32 aKeyCode, PRBool& aHandledFlag)
     if (theDirection == eNavigationDirection_Start) {
       // Close it up.
       mCurrentMenu->OpenMenu(PR_FALSE);
+      NS_ENSURE_TRUE(weakFrame.IsAlive(), NS_OK);
       // SelectMenu() so DOMMenuItemActive is fired for accessibility
       mCurrentMenu->SelectMenu(PR_TRUE);
       aHandledFlag = PR_TRUE;
@@ -1779,6 +1843,7 @@ nsMenuPopupFrame::HideChain()
   
   nsIFrame* frame = GetParent();
   if (frame) {
+    nsWeakFrame weakMenu(frame);
     nsIMenuFrame* menuFrame;
     if (NS_FAILED(CallQueryInterface(frame, &menuFrame))) {
       nsIPopupSetFrame* popupSetFrame = GetPopupSetFrame(mPresContext);
@@ -1789,7 +1854,9 @@ nsMenuPopupFrame::HideChain()
     }
    
     menuFrame->ActivateMenu(PR_FALSE);
+    NS_ENSURE_TRUE(weakMenu.IsAlive(), NS_OK);
     menuFrame->SelectMenu(PR_FALSE);
+    NS_ENSURE_TRUE(weakMenu.IsAlive(), NS_OK);
 
     // Get the parent.
     nsIMenuParent *menuParent = menuFrame->GetMenuParent();
@@ -1954,8 +2021,9 @@ nsMenuPopupFrame::MoveToAttributePosition()
   xPos = left.ToInteger(&err1);
   yPos = top.ToInteger(&err2);
 
-  if (NS_SUCCEEDED(err1) && NS_SUCCEEDED(err2))
-    MoveTo(xPos, yPos);
+  if (NS_SUCCEEDED(err1) && NS_SUCCEEDED(err2)) {
+    MoveToInternal(xPos, yPos);
+  }
 }
 
 
@@ -2105,8 +2173,25 @@ nsMenuPopupFrame::MoveTo(PRInt32 aLeft, PRInt32 aTop)
   left.AppendInt(aLeft);
   top.AppendInt(aTop);
 
+  nsWeakFrame weakFrame(this);
   mContent->SetAttr(kNameSpaceID_None, nsXULAtoms::left, left, PR_FALSE);
+  if (!weakFrame.IsAlive()) {
+    return;
+  }
   mContent->SetAttr(kNameSpaceID_None, nsXULAtoms::top, top, PR_FALSE);
+  if (!weakFrame.IsAlive()) {
+    return;
+  }
+
+  MoveToInternal(aLeft, aTop);
+}
+
+void
+nsMenuPopupFrame::MoveToInternal(PRInt32 aLeft, PRInt32 aTop)
+{
+  // just don't support moving popups for content shells
+  if (mInContentShell)
+    return;
 
   nsIView* view = GetView();
   NS_ASSERTION(view->GetParent(), "Must have parent!");
